/*
 * @(#)DocSet.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.io.*;
import java.util.*;

/**

 * ʸ񽸹Υ饹ʸڤ°(attribute treeˤȤư
 * 롥ǡ°ڤȤϡλҥΡɤΥ٥뤬٤ưۤʤ
 * դڤ򤤤ܼǤϡ̾Υǡڥեɤ߹ȡ
 * ξߤ褦Ԥ
 * <UL>
 * <LI> λҶϡΥ٥ξ˺鱦إȤ롥
 * <LI> ΤƱ٥ĥΡɤܤ˽и硤
 * ʬڤϽ롥
 * </UL>AttrDocSet.java
 * Created: Wed May 18, 2005.
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.13 $
 */
public class AttrDocSet extends DocSet 
{
	/*
	 * Static fields
	 */
    static private boolean DEBUG = false;
    static private boolean DEBUG_MATCH = false;

	/** Removing repeated children */
	public void pruneSiblings() {
		ChildrenSortTreeWalker visitor = new ChildrenSortTreeWalker(this);

		try {
			for (int i = 0; i < this.docList.size(); i++) {
				UnotNode root = (UnotNode) this.docList.get(i);
				
				//ʸڤ󤹤: 
				this.traverseDocTree(root, visitor); 
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}

class ChildrenSortTreeWalker extends TreeWalker {

	NodeLabelComparator comparator;

	ChildrenSortTreeWalker(DocSet docSet) {
		this.comparator = new NodeLabelComparator();
	}

	void work(UnotNode node, UnotNode parent, int depth) throws Exception {
		// ΡɤǻŻ򤹤:
		NodeList list = node.getChildren();

		//٥IDξ˥Ȥ
		//Collections.sort(list, this.comparator);//orig
		Collections.sort(list, this.comparator);

		//٥IDʣǤ
		NodeList newList = new NodeList();
		int last = -1; //undef
		for (int i = 0; i < list.size(); i++) {
			UnotNode child = (UnotNode)list.get(i);
			int label = child.getLabelId();
			if (label != last) {
				newList.add(child);
				last = label;
			}
		}
		node.setChildren(newList);
	}
}

class NodeLabelComparator implements Comparator<UnotNode> {
  
  // 󥹥ȥ饯
  public NodeLabelComparator() {
    super();
  }
  
  public boolean equals(Object obj) {
    return (super.equals(obj));
  }
  
  // ¿˻ꤹ
  //  Integer ֥ȤǤ錄äƤ
  public int compare(UnotNode obj1, UnotNode obj2) {
	  int node1 = ((UnotNode)obj1).getLabelId();
	  int node2 = ((UnotNode)obj2).getLabelId();
	  return (node1 - node2); //
  }
}


class OrgNodeLabelComparator implements Comparator {
  
  // 󥹥ȥ饯
  public OrgNodeLabelComparator() {
    super();
  }
  
  public boolean equals(Object obj) {
    return (super.equals(obj));
  }
  
  // ¿˻ꤹ
  //  Integer ֥ȤǤ錄äƤ
  public int compare(Object obj1, Object obj2) {
	  int node1 = ((UnotNode)obj1).getLabelId();
	  int node2 = ((UnotNode)obj2).getLabelId();
	  return (node1 - node2); //
  }
}


