/*
 * @(#)DLpair.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

/**
 * DLpair.java
 *
 *
 * Created: Thu Aug  7 21:31:44 2003
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.1.1.1 $
 */
public final class DLpair
{

    private int depth;
    private int label;


    /**
     * Creates a new <code>DLpair</code> instance.
     *
     */
    public DLpair()
    {
        // do nothing.
    }


    /**
     * Creates a new <code>DLpair</code> instance.
     *
     * @param depth an <code>int</code> value
     * @param label an <code>int</code> value
     */
    public DLpair(int depth, int label)
    {
        this.depth = depth;
        this.label = label;
    }


    /**
     * Gets the value of depth.
     *
     * @return the value of depth
     */
    public final int getDepth()
    {
        return this.depth;
    }


    /**
     * Gets the value of label.
     *
     * @return the value of label
     */
    public final int getLabel()
    {
        return this.label;
    }


    /**
     * Returns the (depth, label) expression of this object.
     *
     * @return a <code>String</code> value
     */
    public String toString()
    {
        return "(" + Integer.toString(this.depth) + ","
            + Integer.toString(this.label) +")";
    }


    /**
     * Compares this object to the specified object.
     * The result is <code>true</code> if the argument is not
     * <code>null</code> and is a <code>DLpair</code> object that represents
     * the same (depth, label) pair as this object.
     *
     * @param obj an <code>Object</code> value
     * @return a <code>boolean</code> value
     */
    public boolean equals(Object obj)
    {
        if (this == obj) {
            return true;
        }
        if (obj != null && getClass() == obj.getClass()) {
            DLpair pair = (DLpair) obj;
            if (this.depth == pair.depth
                && this.label == pair.label) {
                return true;
            }
        }
        return false;
    }


} // DLpair
