/*
 * @(#)DocSet.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.io.*;
import java.util.*;

/**
 * ʸ񽸹Υ饹ʸ񽸹ϡʸɽ魯Ρɥ٥դڤʤ뿹(a forest)Ǥ롥
 * <UL>
 * <LI> ե  ե̾ġ̾ΥեΤȤؤ
 * <LI> ɥ  `@doc' Ƕڤ줿ǡΤȤؤ
 *      `@doc' Ĥ̵ϡեΤ򣱥ɥȤȤߤʤ
 *      ޤɥȤκǽΥΡɤϥ롼ȥΡɤȤư롥
 *
 * <LI> ΡID  Ρɤ˿դIDΥץϢ³ int ͤͿ롥
 * <LI> Ρֹ  ɥȤ˽񤫤ƤΡɤֹ桥
 *      ܥΡɤɽݤѤ롥ΡIDȤ̵ط
 *      ͤϳƥɥȤȤΩǤꡤ̤ʤΤȤư롥
 *      Ρֹϲ֤ϤޤäƤƤ褤Ϣ³ͤǤäƤ褤
 * </UL>
 *
 * DocSet.java
 * Created: Fri Aug 29 16:15:52 2003
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.14 $
 */
public class DocSet
{
	/*
	 * Static fields
	 */
    static private boolean DEBUG = false;
    static private boolean DEBUG_MATCH = false;

	public static final String CATEGORY_STRING = "@cat";
	public static final String CATEGORY_POSITIVE_STRING = "@pos";
	public static final String CATEGORY_NEGATIVE_STRING = "@neg";
	public static final String STDIN_STRING = "-"; 
	public static final int CATEGORY_POSITIVE_CATID = 0;
	public static final int CATEGORY_NEGATIVE_CATID = 1;

	/** ѥϼ³ڤؤ */
	public static final int READER_PARENT = 0; 
	public static final int READER_CHILD  = 1; 
	public static final int READER_XML    = 2; 
    static int DATA_READER_TYPE    = READER_PARENT;

	/** ܥꥹɽȤɤΥե饰
			ʸߤμǤϡUnot = true, Freqt = false ȤäƤ*/
	public  static boolean useAdjacencyList = true;

	/** ʸ񽸹νϤʲ򤹤뤫ɤΥե饰*/
	public  static boolean useIndent = false;

	/*
	 * Member fields
	 */

	/** ΡID򥭡ΡɤͤȤޥå */
	private NodeDict nodeDict;

	/** ٥ID򥭡٥ʸͤȤޥå */
	private LabelDict labelDict;

	/** ɤ߹ޤ줿Ρɤο */
	private int numberOfNodes = 0;

	/** ɤ߹ޤ줿٥ο */
	private int numberOfLabels = 0;

	/** 饹⤫黲Ȥȡʥ */
	private StreamTokenizer st = null;

	/** ߳ƤɥȤΥ롼ȥΡɤΥΡֹޤǤΥեå */
	private int nodeNumOffset = 0;

	/**********************************************************
	 * ʸκΥꥹ
	 **********************************************************/

	/** ɤ߹ޤ줿ɥȤΥ롼ȥΡɤǼꥹ */
	protected NodeList docList;

	public void addRoot(UnotNode root) {
		this.docList.add(root);
	}

	/** ʸꥹȡʿκΥꥹȡˤ֤ */
	public NodeList getDocList() { return docList; }

	/**********************************************************
	 * The current node information:
	 * These informatios are maintained during the scan of the input file, 
	 * and included in a new node when it is created.
	 **********************************************************/

	/** ƤΡID */
	private int freeNodeId = 0;
	/** ߤΥɥID  = numberOfDocs */
	private int curDocId = -1;
	/** ߤΥƥ꡼ID */
	private int curCatId = 0;

	/** ߤΥɥIDĿʤ */
	public void openNextDoc() { this.curDocId++;  }
	public void openNextDoc(int docId) { this.curDocId = docId;  }

	public int  getNumOfDoc()  { return curDocId + 1; }
	public int  getNumOfCat()  { return curCatId + 1; }

	UnotNode createNewNodeCurrent() {
		UnotNode node = new UnotNode(this.freeNodeId++);
		node.setDocId(this.curDocId);
		node.setCatId(this.curCatId);
		this.numberOfNodes++;
		return node;
	}

	UnotNode createNewNodeCurrent(int labelId) {
		UnotNode node = createNewNodeCurrent();
		node.setLabelId(labelId);
		return node;
	}

	/**
	 * Creates a new <code>DocSet</code> instance.
	 *
	 */
	public DocSet()
	{
		this.docList = new NodeList();
		this.nodeDict = new NodeDict();
		this.labelDict = new LabelDict();
		// (ʲνϾĹ)
	}

	public NodeDict  getNodeDict()  { return nodeDict; }
	public LabelDict getLabelDict() { return labelDict; }
	public void setIndent(boolean flag) { useIndent = flag; }

	/*********************************************************
	 * Input 
	 *********************************************************/

	/**
	 * ե̾ǤȤꥹȤ readFile() ǥɥȤɤ߹ࡥ
	 *
	 * ե̾Ȥ CATEGORY_STRING (= @cat) Ϳ줿硤ƥ꡼
	 * ID 򥤥󥯥Ȥ롥
	 * ƱͤˡCATEGORY_POSITIVE_STRING (= @pos) Ϳ줿祫ƥ꡼ID 
	 * CATEGORY_POSITIVE_CATID ˡ
	 * CATEGORY_NEGATIVE_STRING (= @neg) Ϳ줿祫ƥ꡼ID 
	 * CATEGORY_NEGATIVE_CATID ˥åȤ롥
	 *
	 * @param files a <code>List</code> value
	 */
	public void readFiles(List files)
	{
		for (int i = 0; i < files.size(); i++) {
			String filename = (String) files.get(i);
			if (filename.equals(CATEGORY_STRING)) {
				this.curCatId++;
			} else if (filename.equals(CATEGORY_POSITIVE_STRING)) {
				this.curCatId = CATEGORY_POSITIVE_CATID;
			} else if (filename.equals(CATEGORY_NEGATIVE_STRING)) {
				this.curCatId = CATEGORY_NEGATIVE_CATID;
			} else if (filename.equals(STDIN_STRING)) {
				readFile(System.in);
			} else {
				readFile(filename);
			}
		}
	}

	/**
	 * ̾Τfilename Υեɤ߹ǥǡ򹹿롥
	 *
	 * @param filename a <code>String</code> value
	 * @see file:/usr/share/doc/j2sdk1.4-doc/1.4.1/api/java/io/StreamTokenizer.html
	 */
	public void readFile(String filename)
	{
		try {
			UnotLog.log("@log Read_File: " + filename + "\n");
			readFile(new FileReader(filename));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void readFile(InputStream is)
	{
		UnotLog.log("@log Read_File: InputStream("+is+")\n");
		readFile(new BufferedReader(new InputStreamReader(is)));
	}

	public void readFile(Reader reader)
	{
		// Υɥɤ߹Ρɤο
		int numNodeReadInDoc = 0;

		try {

			// tokenizer 
			st = new StreamTokenizer(reader);
			//st = new StreamTokenizer(new FileReader(filename));
			st.eolIsSignificant(true);
			st.slashSlashComments(true);
			st.wordChars('@', '@'); //@ñʸȤưʤ

			// lineReader 
			// ParentFormatLineReader or ChildFormatLineReader is available.
			LineReader lineReader = new ParentFormatLineReader(this, st);

			while (st.nextToken() != StreamTokenizer.TT_EOF) {

				// ʸȽ
				if (st.ttype == StreamTokenizer.TT_WORD) {

					//ԤƬʬ: 
					if (st.sval.equals("@doc") ||
						st.sval.equals("@pattern")) {

						//Initialization of Ρɤ̾
						openNextDoc(); 
						numNodeReadInDoc = 0;
						this.nodeDict.makeMapEmpty(); //ΡID֤̾򿷵ˤ롥

						lineReader.skipRestOfLine(); //ԤλĤɤФ
					}
					else {
						lineReader.skipRestOfLine(); //ԤλĤɤФ
					}
					
				} else if (st.ttype == StreamTokenizer.TT_NUMBER) {
					// ΡɹԤɤ߹

					if (this.curDocId < 0) //̤ʾ硧Ƭ@docʤȤ
						openNextDoc(); 

					UnotNode node = lineReader.readNodeLine();

					if (numNodeReadInDoc == 0)
						addRoot(node); //ʸκǽΥΡɤ򺬤ȤϿ롥

						//this.docList.add(node);
					numNodeReadInDoc++;
					DocSet.this.numberOfNodes++;
				}
			}

			UnotLog.log("@log NumOfDataNodes " +this.numberOfNodes+ "\n");
			UnotLog.log("@log NumOfDataDocs  " +this.getNumOfDoc()+ "\n");
			UnotLog.log("@log NumOfDataCats: " +this.getNumOfCat()+ "\n");

			return;

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * ̾Τfilename Υեɤ߹ǥǡ򹹿롥
	 *
	 * @param filename a <code>String</code> value
	 * @see file:/usr/share/doc/j2sdk1.4-doc/1.4.1/api/java/io/StreamTokenizer.html
	 */
	public void old_readFile(String filename)
	{
		UnotLog.log("@log Read_File: " + filename + "\n");


		// Υɥɤ߹Ρɤο
		int numNodeReadInDoc = 0;

		try {

			// tokenizer 
			st = new StreamTokenizer(new FileReader(filename));
			st.eolIsSignificant(true);
			st.slashSlashComments(true);
			st.wordChars('@', '@'); //@ñʸȤưʤ

			// lineReader 
			// ParentFormatLineReader or ChildFormatLineReader is available.
			LineReader lineReader = new ParentFormatLineReader(this, st);

			while (st.nextToken() != StreamTokenizer.TT_EOF) {

				// ʸȽ
				if (st.ttype == StreamTokenizer.TT_WORD) {

					//ԤƬʬ: 
					if (st.sval.equals("@doc") ||
						st.sval.equals("@pattern")) {

						//Initialization of Ρɤ̾
						openNextDoc(); 
						numNodeReadInDoc = 0;
						this.nodeDict.makeMapEmpty(); //ΡID֤̾򿷵ˤ롥

						lineReader.skipRestOfLine(); //ԤλĤɤФ
					}
					else {
						lineReader.skipRestOfLine(); //ԤλĤɤФ
					}
					
				} else if (st.ttype == StreamTokenizer.TT_NUMBER) {
					// ΡɹԤɤ߹

					if (this.curDocId < 0) //̤ʾ硧Ƭ@docʤȤ
						openNextDoc(); 

					UnotNode node = lineReader.readNodeLine();

					if (numNodeReadInDoc == 0)
						addRoot(node); //ʸκǽΥΡɤ򺬤ȤϿ롥

						//this.docList.add(node);
					numNodeReadInDoc++;
					DocSet.this.numberOfNodes++;
				}
			}

			UnotLog.log("@log NumOfDataNodes " +this.numberOfNodes+ "\n");
			UnotLog.log("@log NumOfDataDocs  " +this.getNumOfDoc()+ "\n");
			UnotLog.log("@log NumOfDataCats: " +this.getNumOfCat()+ "\n");

			return;

		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/*********************************************************
	 * Occurrence List 
	 *********************************************************/

    /**
     * ǾݡȤѡɽниѴ롥
     */
    public int convertMinSupRatioToAbsolute(double minSupRatio)
    {
        int minsup = (int) Math.ceil(this.getDataSize() * minSupRatio);
		if (minsup <= 1) {
			minsup = Math.max(minsup, 2); 
			UnotLog.log("@log Warning: Minimum_Support_Absolute must be >= 2. "+
						"Automatically changed to: " + minsup + "\n");
		}
        return minsup; 
    }


	/** ߴΤΥåѡ³ */
	public OccListTable getOccLists() { return getOccLists(this.docList); }

	/** ߴΤΥåѡ³ */
	public OccListTable getOccLists(int minSup) { return getOccLists(); }

	/**
	 * ƥ٥IDȤνиꥹȤ򡤳ƥ٥IDźȤݻɽ
	 */
	public OccListTable getOccLists(NodeList docList)
	{
		int maxLabelSize = this.labelDict.labelMap.size();
		OccListTable occLists = new OccListTable(maxLabelSize);
		MultiNodeGatherer visitor = new MultiNodeGatherer(occLists);

		if (DEBUG_MATCH)
			System.out.println("@DocSet docList: "+
							   this.getDocList().getNodeIdListString());

		try {
			for (int i = 0; i < docList.size(); i++) {
				UnotNode root = (UnotNode) docList.get(i);
				if (DEBUG_MATCH)
					System.out.println("@DocSet get root: id="+root.getNodeId());
				
				//ʸڤ󤹤: ˬ䤷 occLists ɲä롥
				this.traverseDocTree(root, visitor);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return occLists;
	}

	public NodeList getANodeList(NodeList rootList, int labelId) {
		NodeList nodeList = new NodeList();
		NodeGathererNodeList visitor = new NodeGathererNodeList(nodeList, labelId);

		try {
			for (int i = 0; i < rootList.size(); i++) {
				//ʸڤ󤹤: ˬ䤷 nodeLists ɲä롥
				UnotNode root = (UnotNode)rootList.get(i);
				this.traverseDocTree(root, visitor);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return nodeList;
	}

	/**
	 * Υǡγƥ٥IDȤνиꥹΤνޤ֤
	 * νޤϡƥ٥IDźˤ
	 * б륨ȥˤΥ٥ID٥ˤĥΡɤΥꥹȤġ
	 *
	 * @return a list <code>List</code> of the root occurrence
	 * list for all single node patterns. 
	 */
	public OccListTable getOccLists(NodeDict nodeDict)
	{
		int maxLabelSize = this.labelDict.labelMap.size();
		OccListTable occLists = new OccListTable(maxLabelSize);

		// list  occLists γ OccList Ϣ
		NodeList list = this.nodeDict.nodeList; 
		for (int i = 0; i < list.size(); i++) {
			UnotNode node = (UnotNode) list.get(i);
			if (node == null) continue;

			// Ρɤ򤽤Υ٥˱ɽɲ
			occLists.putOcc(node.getLabelId(),node);
		}
		return occLists;
	}

	/**
	 * ǡΥ֤
	 * ϥɥȤʣξ硤ɥȿ֤
	 * ϥɥȤñξ硤Ρɿ֤
	 *
	 * @return an <code>int</code> value
	 */
	public final int getDataSize()
	{
		if (this.curDocId > 1) {
			return this.getNumOfDoc();
		} else {
			return this.numberOfNodes;
		}
	}


	/**
	 * Gets the number of doc/root occurrence.
	 * If the number of traversed documents is only one,
	 * returns the number of root occurrence.
	 * Otherwise returns the number of document occurrence.
	 *
	 * @return an <code>int</code> value
	 */
	public final int getOccSize(OccList occList)
	{
		if (this.getNumOfDoc() > 1) {
			return occList.getDocCount();
		} else {
			return occList.getRootCount();
		}
	}


	/**
	 * Describe <code>setAdjacencyListRepl</code> method here.
	 *
	 * @param bool a <code>boolean</code> value
	 */
	public static void useAdjacencyList(boolean bool)
	{
		DocSet.useAdjacencyList = bool;
	}

	/*********************************************************
	 * Dictionaries 
	 * 
	 *********************************************************/

	/* Added by arim for label string output */
	/**
	 * ٥ID餽Υ٥ʸ֤
	 * ٥ʸŰǰϤޤ롥
	 */
	public String lookupLabelById(int id) {
		return "\""+this.labelDict.lookupLabel(id)+"\"";
	}

	/**
	 * Ρ ̾ȥΡɤбդ뤿μ
	 * Ρ̾0ʾǤꡤեѤ롥äˡ
	 Ρ̾-1פФ̤ʸΡɤβŪʿƥΡɤȤ
	 * nullбդ(the master root)
	 * աƥΡ̾̾פʸǤꡤ
	 * ƥΡID̾פϡĤDocSetǤ뤳Ȥա
	 *
	 * @author <a href="mailto:arim@ist.hokudai.ac.jp">Hiroki Arimura</a>
	 * @version 1.0
	 */
	public class NodeDict
	{

		/** Ρɤľ˳ǼꥹȡΡID򥭡Ȥ
				Ρɤ򻲾Ȥ뤿μȤѤ롥 */
		private NodeList nodeList;

		/** Ρɤľ˳ǼꥹȡΡID򥭡Ȥ
				Ρɤ򻲾Ȥ뤿μȤѤ롥 */
		//private Map nameMap;
		private Map<Integer,UnotNode> nameMap;

		/**
		 * Creates a new <code>NodeDict</code> instance.
		 *
		 */
		public NodeDict()
		{
			this.nodeList = new NodeList();
			//this.nameMap  = new HashMap();
			this.nameMap  = new HashMap<Integer,UnotNode>();
			DocSet.this.numberOfNodes = 0;
		}

		public void makeMapEmpty() {
			this.nameMap = null; //ϡMapΤƤ뤳Ȥ뤿Τߡ
			//this.nameMap = new HashMap();
			this.nameMap = new HashMap<Integer,UnotNode>();
		}

		/** ΡɥꥹȤ֤ꥹΥΡɤϡ
		 *  줿¤Ǥ뤳Ȥݾ㤵롥*/
		public NodeList getNodeList() { return nodeList; }

		/**
		 * Ρ ID ФơбΡɤ֤
		 * ΥΡ ID ĥΡɤ¸ߤʤäˤϡ
		 * ˤΤ褦ʥΡɤƤ֤
		 * ID -1 ФƤϡnull֤
		 *
		 * @param id an <code>int</code> value
		 * @return an <code>UnotNode</code> value
		 */
		protected UnotNode lookupNode(int id)
		{
			if (id < 0)
				//ID=-1ФƥޥΡɤɽ魯 null ֤
				return null;

			Integer ID = new Integer(id);
			
			if (this.nameMap.containsKey(ID)) {
				// ΥΡID̾ϴ¸ߤƤΤǡбΡɤ֤

				return (UnotNode) this.nameMap.get(ID);
								
			} else {

				// ΥΡID̾¸ߤʤΤǡ˥Ρɤ֤
				UnotNode node = DocSet.this.createNewNodeCurrent(); 

				this.nodeList.add(node);
				this.nameMap.put(ID, node);
				return node;
			}
		}

	} // NodeDict

	/**
	 * Ρ ID ˥Ρɤбդ뤿μ
	 * ΡID0ʾǤꡤID -1 Ф null Ρɤбդ롥
	 * ID -1 ΥΡɤϤ٤ƤʸΡɤβŪʿƥΡɤɽ
	 * ʥΡɤΡ ID 򻲾ȤˤϡUnotNode.getNodeId()Ѥ.
	 *
	 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
	 * @version 1.0
	 */
	public class OldNodeDict
	{

		/** Ρɤľ˳ǼꥹȡΡID򥭡Ȥ
				Ρɤ򻲾Ȥ뤿μȤѤ롥 */
		private NodeList nodeList;

		/** ƤΡID */
		private int freeNodeId;


		/**
		 * Creates a new <code>NodeDict</code> instance.
		 *
		 */
		public OldNodeDict()
		{
			this.nodeList = new NodeList();
			this.freeNodeId = 0;
			DocSet.this.numberOfNodes = 0;
		}

		/** ΡɥꥹȤ֤ꥹΥΡɤϡ
		 *  줿¤Ǥ뤳Ȥݾ㤵롥*/
		public NodeList getNodeList() { return nodeList; }

		/**
		 * Ρ ID ФơбΡɤ֤
		 * ΥΡ ID ĥΡɤ¸ߤʤä硤
		 * ˤΤ褦ʥΡɤƤ֤
		 * ID -1 ФƤϡnull֤
		 *
		 * @param id an <code>int</code> value
		 * @return an <code>UnotNode</code> value
		 */
		protected UnotNode lookupNode(int id)
		{
			// id ޤǤʤ null 
			// ( get(id) Ԥݤ IndexOutOfBoundsException ȯʤ
			if (id < 0)
				//ID=-1Ф null ֤
				return null;
			
			for (int i = this.nodeList.size(); i <= id; i++) {
				this.nodeList.add(null);
			}

			if (this.nodeList.get(id) != null) {
				// ˺ƤΤǤ֤
				return (UnotNode) this.nodeList.get(id);

			} else {

				// null ƤʤΤǡnode ­֤
				UnotNode node = DocSet.this.createNewNodeCurrent(); 
				//	UnotNode node = new UnotNode(DocSet.this.curDocId, this.freeNodeId);
				//	this.freeNodeId++;
				//	node.setCatId(DocSet.this.curCatId);
				this.nodeList.set(id, node);
				return node;

			}
		}

	} // NodeDict


    /**
     * ٥ʸ <=> ٥ ID ߻Ȥ뤿μ
     *
     * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
     * @version 1.0
     */
	public class LabelDict
	{

		/** ٥ʸ򥭡٥ ID ͤȤޥå */
		//protected Map labelMap;
		protected Map<String,Integer> labelMap;

		/** ٥ ID 򥭡٥ʸͤȤޥå */
		protected Map<Integer,String> idMap; // ArrayList ǼƤ褤

		/** Ƥ٥ID */
		private int freeLabelId;

		/**
		 * Creates a new <code>LabelDict</code> instance.
		 *
		 */
		public LabelDict()
		{
			//this.labelMap = new HashMap();
			this.labelMap = new HashMap<String,Integer>();
			this.idMap = new HashMap<Integer,String>();
			this.freeLabelId = 0;
			DocSet.this.numberOfLabels = 0;
		}


		/**
		 * ٥ʸ󤫤٥ ID ֤
		 * Τ褦ʥ٥뤬ϿƤʤСΥ٥ʸ򥭡
		 * ʥ٥ ID ͤȤƼ򹹿롥
		 *
		 * @param label a <code>String</code> value
		 * @return an <code>int</code> value
		 */
		public int lookupId(String label)
		{
			if (this.labelMap.containsKey(label)) {

				// Υ٥̾ϴ¸ߤƤΤǤΤޤ ID ֤

				Integer id = (Integer) this.labelMap.get(label);
				return id.intValue();
								
			} else {

				// ƽи٥̾ʤΤǡfreeLabelId ȥޥåפ򹹿

				Integer id = new Integer(this.freeLabelId);
				this.freeLabelId++;
				DocSet.this.numberOfLabels++;

				this.labelMap.put(label, id);
				this.idMap.put(id, label);

				//  ID ֤
				return id.intValue();
			}
		}


		/**
		 * ٥ ID ٥ʸ֤
		 *
		 * @param id an <code>int</code> value
		 * @return a <code>String</code> value
		 */
		public String lookupLabel(int id)
		{
			Integer _id = new Integer(id);

			if (this.idMap.containsKey(_id)) {
				//modified by arim@ist on 041220
				return (String) this.idMap.get(_id); 
				//return (String) this.labelMap.get(_id);//error

			} else {
				System.exit(Unot.EXIT_FAILURE);
				return null;
			}
		}

	} // LabelDict


	//See ChildFormatLineReader.java

	/*********************************************************
	 * Output
	 *********************************************************/

	/**
	 * ʸ񽸹祪֥Ȥ򡤺ɤ߹߲ǽʷǥե˽Ϥ롥
	 *
     * @author Hiroki Arimura
	 */
	public void writeFile(PrintStream out)
	{

		NodeList docList = this.docList;
		UnotNode parent  = null;

		try {

			//
			LineWriter lineWriter = new ParentFormatLineWriter(this, out);
			lineWriter.setIndent(DocSet.useIndent);
			WriteFileTreeWalker visitor = new WriteFileTreeWalker(lineWriter);
		
			if (DEBUG) System.out.println("debug: writeFile: enter");
			out.println("@writer "+lineWriter.getName());

			// ʸڤ򤽤κ󤹤롥
			for (int i = 0; i < docList.size(); i++) {
				UnotNode root = (UnotNode) docList.get(i);
				if (root == null) continue; //skip for a null node
				lineWriter.writeDocDelimitor(root.getDocId());

				//ʸڤν
				this.traverseDocTree(root, visitor);
			}
			if (DEBUG) System.out.println("debug: writeFile: exit");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/** ΡɤΥ٥ʸ֤ƥƥؿ
	 */
	public String getLabelString(UnotNode node) {
		if (node == null)
			return "***null***";
		//throw new IOException();

		return this.lookupLabelById(node.getLabelId());
	}


	/*********************************************************
	 * TreeWalker 
	 *********************************************************/

	/** ڽ³ǡڤͿ줿Ρɤ򺬤Ȥơ
	 *  βʬڤ󤷤ƺȤ롥
	 *  ˬ䤷ƥΡɤǡvisitor (node, parent, depth)ȤƼ¹Ԥ롥
	 */
	public void traverseDocTree(UnotNode root, TreeWalker visitor) throws Exception {
		traverseDocTree(root, null, 0, visitor);
	}

	private void traverseDocTree(UnotNode node, UnotNode parent, int depth,
								 TreeWalker visitor) throws Exception 
	{
		if (node == null)
			return; 

		// ΡɤǻŻ򤹤
		visitor.work(node, parent, depth);

		parent = node;
		NodeList children = node.getChildren();
		for (int i = 0; i < children.size(); i++) {
			UnotNode child = (UnotNode) children.get(i);
			traverseDocTree(child, node, depth+1, visitor);
		}
	}

	/*********************************************************
	 * Main
	 *********************************************************/


    /**
     * ƥ
     *
     * @param args[] a <code>String</code> value
     */
	public static void main(String args[])
	{
		if (args.length < 1) {
			System.out.println("USAGE: java unot.DocSet FILENAME");
			return;
		}

		String filename = args[0];
		DocSet docset = new DocSet();
		docset.readFile(filename);
		System.out.println("numberOfDocs = "+docset.getNumOfDoc());
		System.out.println("numberOfNodes = "+docset.numberOfNodes);
		docset.writeFile(System.out);
		return;
	}

} // DocSet


/** ǡڤΥΡɤ󤹤ԤΥƥץ졼ȡ */
abstract class TreeWalker {
	abstract void work(UnotNode node, UnotNode parent, int depth) throws Exception;
}

class WriteFileTreeWalker extends TreeWalker {
	LineWriter lineWriter = null;

	WriteFileTreeWalker(LineWriter lineWriter) {
		this.lineWriter = lineWriter; 
	}

	void work(UnotNode node, UnotNode parent, int depth) throws IOException {
		// ΡɤǻŻ򤹤
		lineWriter.putIndent(depth);
		lineWriter.writeNodeLine(node);
	}
}

class MultiNodeGatherer extends TreeWalker {
    static private boolean DEBUG_MATCH = false;

	OccListTable occLists = null;

	MultiNodeGatherer(OccListTable occLists) {
		this.occLists = occLists; 
	}

	void work(UnotNode node, UnotNode parent, int depth) throws Exception {
		// ΡɤǻŻ򤹤: 
		// Ρɤ򤽤Υ٥˱ɽɲ
		if (DEBUG_MATCH) {
			System.out.println("@DocSet --visit node:"+
							   " id="+node.getNodeId()+
							   " label="+node.getLabelId());
			OccList occList =
				this.occLists.getOccList(node.getLabelId()); 
			System.out.println("        before -- new OccList["+node.getLabelId()+"]: "+
							   occList.getRootList().getNodeIdListString());
		}

		this.occLists.putOcc(node.getLabelId(),node);

		if (DEBUG_MATCH) {
			OccList occList =
				this.occLists.getOccList(node.getLabelId()); 
			System.out.println("        after  -- new OccList["+node.getLabelId()+"]: "+
							   occList.getRootList().getNodeIdListString());
		}

	}
}

class NodeGathererNodeList extends TreeWalker {
	NodeList nodeList = null;
	int labelId = -1;

	NodeGathererNodeList(NodeList nodeList, int labelId) {
		this.nodeList = nodeList; 
		this.labelId = labelId; 
	}

	void work(UnotNode node, UnotNode parent, int depth) throws Exception {
		// ΡɤǻŻ򤹤: 
		// Ρɤ򤽤Υ٥˱ɽɲ
		if (node.getLabelId() == labelId)
			this.nodeList.add(node);
	}
}

