/*
 * @(#)Occurrence.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This file contains an implementation of the algorithm "UNOT".
 *
 * @see Discovering Frequent Substructures in Large Unordered Trees,
 * In Proc. the 6th Discovery Science (DS'03), LNAI, October 2003.
 * <http://www.i.kyushu-u.ac.jp/doitr/trcs216.pdf>
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */
package unot;

import java.util.*;

/**
 * Occurrence.java
 *
 * ǡˤѥνиɽ魯
 * NΥѥ(0,...,N-1)
 * νиfϡ٥ȥ(f(0),...,f(N-1))ɽ롥
 * ٤ƤνиΤϡ̤ƬͭƬڤȤƴ롥
 * ƥ֥ȤϡĹNνи٥ȥб륻ɽ魯
 * ܺ١٥ȥϡԼԤ򤵤եݥprevϢ뤷ꥹȤǤ롥
 * ƥϡ٥ȥκǽΥʥ롼Ȥνи򤵤ˤ򤵤ݥfirstġˤꡤ֤ǥ롼Ƚи֤Ф롥
 * Created: Mon Dec 29 03:51:13 2003
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.3 $
 */
public class Occurrence
{

    /**  */
    private int size;

    /**  */
    private Occurrence prev;

    /**  */
    private Occurrence first;

    /**  */
    private UnotNode element;

    /**  */
    private transient Occurrence next;


    /**
     * Creates a new <code>Occurrence</code> instance.
     *
     */
    public Occurrence()
    {
        this.size = 0;
        this.prev = null;
        this.first = this;
        this.element = null;
    }

    /**
     * Creates a new <code>Occurrence</code> instance.
     *
     * @param occ an <code>Occurrence</code> value
     */
    public Occurrence(Occurrence occ)
    {
        this.size = occ.size;
        this.prev = occ.prev;
        this.first = occ.first;
        this.element = occ.element;
    }


    /**
     * ĹN-1νи٥ȥɽ魯ˡ
	 * NܤΥѥνиΡchild
	 * ɲäơĹNνи٥ȥɽ魯֤
     * @return ĹNνи٥ȥɽ魯롥
     */
    public Occurrence(Occurrence occ, UnotNode child)
    {
        this.size = occ.size + 1;
        this.prev = occ;
        this.first = occ.first;
        this.element = child;
    }


    /**
     * Describe <code>add</code> method here.
     *
     * @param o an <code>UnotNode</code> value
     * @return a <code>boolean</code> value
     */
    public boolean add(UnotNode o)
    {
        if (this.element == null) {
            this.size++;
            this.element = o;
        } else {
            Occurrence occ = new Occurrence(this);
            this.size++;
            this.prev = occ;
            this.element = o;
        }
        return true;
    }


    /**
     * Describe <code>contains</code> method here.
     *
     * @param o an <code>UnotNode</code> value
     * @return a <code>boolean</code> value
     */
    public boolean contains(UnotNode o)
    {
        Occurrence occ = this;
        while (occ != null) {
            if (occ.element == o) {
                return true;
            } else {
                occ = occ.prev;
            }
        }
        return false;
    }

	public UnotNode getRootOcc() {
		return this.first.element;
	}

    /**
     * Describe <code>get</code> method here.
     *
     * @param index an <code>int</code> value
     * @return an <code>UnotNode</code> value
     */
    public UnotNode get(int index)
    {
        Occurrence occ = this;
        if (this.size <= index) {
            throw new NoSuchElementException();
        }

        if (index == 0) {
            return this.first.element;
        }

        for (int i = index+1; i != this.size; i++) {
            occ = occ.prev;
        }
        return occ.element;
    }

    /**
     * Returns the number of elements in this list.
     *
     * @return an <code>int</code> value
     */
    public int size()
    {
        return this.size;
    }


    /**
     * Retrieve a tree like String representation of the object.
     *
     * @return a <code>String</code> value
     */
    public String toString()
    {
        StringBuffer strBuf = new StringBuffer();
        strBuf.append("docid");
        strBuf.append(((UnotNode) get(0)).getDocId());

        strBuf.append("[");

        Occurrence occ = this;
        occ.next = null;
        while (occ != null) {
            if (occ.prev != null) {
                occ.prev.next = occ;
            }
            occ = occ.prev;
        }

        occ = this.first;
        while (occ != null) {
            if (occ != this.first) {
                strBuf.append(", ");
            }
            strBuf.append(occ.element.toString());
            occ = occ.next;
        }

        strBuf.append("]");
        return strBuf.toString();
    }

} // Occurrence (Trie version)



// // ArrayList version...

// public class Occurrence extends ArrayList
// {

//     public Occurrence(Occurrence occ, NodeAdapter child)
//     {
//         super.addAll(occ);
//         super.add(child);
//         return this;
//     }

//     /**
//      * Retrieve a tree like String representation of the object.
//      *
//      * @return a <code>String</code> value
//      */
//     public String toString()
//     {
//         StringBuffer strBuf = new StringBuffer();
//         strBuf.append(((NodeAdapter) super.get(0)).getDocNumber());
//         strBuf.append("<");
//         strBuf.append(((NodeAdapter) super.get(0)).getDocName());
//         strBuf.append(">: ");
//         strBuf.append(super.toString());
//         strBuf.append("\n");
//         return strBuf.toString();
//     }

// } // Occurrence
